#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Retrievers for RAG.
"""
from abc import ABC, abstractmethod
import copy
import csv
import gzip
import numpy as np
import os
from parlai.core.message import Message
import torch
import torch.cuda
import torch.nn
from tqdm import tqdm

try:
    from transformers import BertTokenizerFast as BertTokenizer
except ImportError:
    from transformers import BertTokenizer
from typing import Tuple, List, Dict, Union, Optional, Any
from typing_extensions import final
from sklearn.feature_extraction.text import TfidfVectorizer

from parlai.agents.tfidf_retriever.tfidf_retriever import TfidfRetrieverAgent
from parlai.core.agents import create_agent, create_agent_from_model_file
from parlai.core.build_data import modelzoo_path
from parlai.core.dict import DictionaryAgent
from parlai.core.loader import register_agent
from parlai.core.opt import Opt
from parlai.core.torch_generator_agent import TorchGeneratorAgent
from parlai.core.torch_ranker_agent import TorchRankerAgent
import parlai.utils.logging as logging
from parlai.utils.torch import padded_tensor
from parlai.utils.typing import TShared

from parlai.agents.rag.dpr import DprQueryEncoder
from parlai.agents.rag.polyfaiss import RagDropoutPolyWrapper
from parlai.agents.rag.indexers import DenseHNSWFlatIndexer, indexer_factory
from parlai.agents.rag.args import (
    RetrieverType,
    WOW_INDEX_PATH,
    WOW_PASSAGES_PATH,
    POLYENCODER_OPT_KEYS,
    TRANSFORMER_RANKER_BASE_OPT,
    WOW_COMPRESSED_INDEX_PATH,
)
from parlai.agents.rag.retrieve_api import SearchEngineRetriever


def load_passage_reader(
    ctx_file: str, return_dict: bool = True
) -> Union[Dict[str, Tuple[str, str]], List[Tuple[str, str, str]]]:
    """
    Load passages from file, corresponding to a FAISS index.

    We attempt to read the passages with a csv reader.

    If passage files are not saved correctly with a csv reader,
    reads can fail.

    :param ctxt_file:
        file to read

    :return reader:
        return a reader over the passages
    """
    logging.info(f'Reading data from: {ctx_file}')
    f_open = gzip.open if ctx_file.endswith(".gz") else open
    try:
        passages = {} if return_dict else []
        with f_open(ctx_file) as tsvfile:
            _reader = csv.reader(tsvfile, delimiter='\t')  # type: ignore
            ids = []
            for idx, row in tqdm(enumerate(_reader)):
                if idx == 0:
                    assert row[0] == 'id'
                    ids.append(-1)
                elif idx <= 1:
                    ids.append(row[0])
                    if return_dict:
                        passages[row[0]] = (row[1], row[2])  # type: ignore
                    else:
                        passages.append((row[0], row[1], row[2]))  # type: ignore
                    continue
                else:
                    assert int(row[0]) == int(ids[idx - 1]) + 1, "invalid load"
                    if return_dict:
                        passages[row[0]] = (row[1], row[2])  # type: ignore
                    else:
                        passages.append((row[0], row[1], row[2]))  # type: ignore
                    ids.append(row[0])

        del ids
    except (csv.Error, AssertionError) as e:
        passages = {} if return_dict else []
        logging.error(f'Exception: {e}')
        logging.warning('Error in loading csv; loading via readlines')
        with f_open(ctx_file) as tsvfile:
            for idx, l in tqdm(enumerate(tsvfile.readlines())):
                line = l.replace('\n', '').split('\t')  # type: ignore
                assert len(line) == 3
                if idx == 0:
                    assert line[0] == 'id'
                if line[0] != 'id':
                    if return_dict:
                        passages[row[0]] = (row[1], row[2])  # type: ignore
                    else:
                        passages.append((row[0], row[1], row[2]))  # type: ignore
    return passages


def load_passages_dict(ctx_file: str) -> Dict[str, Tuple[str, str]]:
    """
    Load passages as a dict.

    :param ctx_file:
        file to read

    :return passages_dict:
        return a dict mapping passage id to a tuple of (text, title)
    """
    psgs_dict = load_passage_reader(ctx_file, return_dict=True)
    assert isinstance(psgs_dict, dict)
    return psgs_dict


def load_passages_list(ctx_file: str) -> List[Tuple[str, str, str]]:
    """
    Load passages as a list.

    :param ctx_file:
        file to read

    :return passages_dict:
        return a list of 3-tuples (id, text, title)
    """
    psgs_list = load_passage_reader(ctx_file, return_dict=False)
    assert isinstance(psgs_list, list)
    return psgs_list


class Document:
    """
    A Document used in retrieval.
    """

    TITLE_DELIM = ' / '
    PASSAGE_DELIM = ' // '

    def __init__(self, title: str, text: str, docid: Union[int, str]):
        assert all(isinstance(t, str) for t in [title, text])
        self._title = title
        self._text = text
        self._id = str(docid)

    def get_title(self) -> str:
        return self._title

    def get_text(self) -> str:
        return self._text

    def get_id(self) -> str:
        return self._id

    def __repr__(self):
        return f"ID: {self._id}\nTitle: {self._title}\nText: {self._text}"

    def __str__(self):
        return f"{self._title} | {self._text}"

    def get_passage_str(self):
        return f"{self._title.strip()}{self.TITLE_DELIM}{self._text.strip()}{self.PASSAGE_DELIM}"

    def get_tokenization_str(self):
        return f"{self._title.strip()}{self.TITLE_DELIM}{self._text.strip()}"


BLANK_DOC = Document('', '', '')


def argsort_scores_and_docs(
    scores: torch.Tensor, docs: List[Document], n_docs: int
) -> Tuple[List[Document], torch.Tensor]:
    """
    Sort scores and documents by score, return n_docs ranked docs/scores.

    :param scores:
        scores with which to rank
    :param docs:
        docs to argsort
    :param n_docs:
        number of docs to return

    :return:
        (docs, scores) --> sorted documents, according to scores.
    """
    scores_sorter = scores.sort(descending=True)
    ranked_docs = [docs[idx] for idx in scores_sorter.indices[:n_docs]]
    ranked_scores = scores_sorter.values[:n_docs]
    return ranked_docs, ranked_scores


def clean_vec(
    vec: torch.LongTensor, end_idx: int, special_toks: List[int] = None
) -> List[int]:
    """
    Remove special tokens from a tensor prior to text conversion.
    """
    new_vec = []
    for i in vec:
        if i == end_idx:
            break
        elif special_toks and i in special_toks:
            continue
        new_vec.append(i)
    return new_vec


class RagRetrieverTokenizer:
    """
    Wrapper for various tokenizers used by RAG Query Model.
    """

    def __init__(
        self,
        query_model: str,
        dictionary: DictionaryAgent,
        max_length: int = 256,
        delimiter='\n',
    ):
        """
        :param query_model:
            query model type (e.g. bert)
        :param dictionary:
            ParlAI dictionary agent
        :param fast:
            whether to instantiate fast BertTokenizer
        :param max_length:
            maximum length of encoding.
        """
        self.query_model = query_model
        self.tokenizer = self._init_tokenizer(dictionary)
        self.max_length = max_length
        self._delimiter = delimiter

    def _init_tokenizer(
        self, dictionary: DictionaryAgent
    ) -> Union[BertTokenizer, DictionaryAgent]:
        """
        If a regular parlai model, use the regular dictionary.

        Otherwise, build as necessary

        :param dictionary:
            ParlAI dictionary agent
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            return BertTokenizer.from_pretrained('bert-base-uncased')
        else:
            return dictionary

    def get_pad_idx(self) -> int:
        """
        Return pad token idx.
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            return self.tokenizer.pad_token_id
        else:
            return self.tokenizer[self.tokenizer.null_token]

    def get_delimiter(self) -> str:
        """
        Return delimiter.
        """
        return self._delimiter

    def get_bos_idx(self) -> int:
        """
        Return start token idx.
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            return self.tokenizer.bos_token_id or 1
        else:
            return self.tokenizer[self.tokenizer.start_token]

    def get_eos_idx(self) -> int:
        """
        Return start token idx.
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            return self.tokenizer.eos_token_id or 2
        else:
            return self.tokenizer[self.tokenizer.end_token]

    def encode(self, txt: str, txt_pair: Optional[str] = None) -> List[int]:
        """
        Encode text.

        :param txt:
            text to encode
        :param txt_pair:
            Optional additional text to encode.
            Useful if encoding two parts of a text, e.g. title & text.

        :return encoding:
            return encoded text.
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            txt = txt.lower().strip()
            if txt_pair:
                txt_pair = txt_pair.lower().strip()
            return self.tokenizer.encode(
                txt,
                text_pair=txt_pair,
                add_special_tokens=True,
                max_length=self.max_length,
                pad_to_max_length=False,
                truncation='longest_first',
            )
        else:
            return self.tokenizer.txt2vec(txt)

    def decode(self, vec: torch.LongTensor) -> str:
        """
        Decode a token vector into a string.
        """
        if self.query_model in ['bert', 'bert_from_parlai_rag']:
            return self.tokenizer.decode(
                clean_vec(vec, self.get_eos_idx()), skip_special_tokens=True
            )
        else:
            return self.tokenizer.vec2txt(
                clean_vec(
                    vec,
                    self.get_eos_idx(),
                    special_toks=[
                        self.get_pad_idx(),
                        self.get_bos_idx(),
                        self.get_eos_idx(),
                    ],
                )
            )


class RagRetriever(torch.nn.Module, ABC):
    """
    RAG Retriever.

    Provides an interface to the RagModel for retrieving documents.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        super().__init__()
        self.retriever_type = RetrieverType(opt['rag_retriever_type'])
        if not (
            (self.retriever_type == RetrieverType.SEARCH_ENGINE)
            or (opt.get('retriever_debug_index') in [None, 'none'])
        ):
            if opt.get('retriever_debug_index') == 'exact':
                opt['path_to_index'] = WOW_INDEX_PATH
            else:
                opt['path_to_index'] = WOW_COMPRESSED_INDEX_PATH
            opt['path_to_dpr_passages'] = WOW_PASSAGES_PATH
        self.opt = opt
        self.print_docs = opt.get('print_docs', False)
        self.max_doc_len = opt['max_doc_token_length']
        self.max_query_len = opt['rag_query_truncate'] or 1024
        self.end_idx = dictionary[dictionary.end_token]
        self._tokenizer = RagRetrieverTokenizer(
            query_model=opt['query_model'],
            dictionary=dictionary,
            delimiter=opt.get('delimiter', '\n') or '\n',
        )
        self.fp16 = (
            not opt['no_cuda']
            and torch.cuda.is_available()
            and self.opt.get('fp16', False)
        )

    @final
    def retrieve(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieve documents, given a query vector.

        :param query:
            tokenized query

        :return (docs, scores):
            docs: list of Documents for each batch example.
            scores: [bsz, n_docs] document scores
        """
        docs, scores = self.retrieve_and_score(query)
        if self.print_docs:
            self.display_docs(docs)
        self.top_docs = [[str(d) for d in ds] for ds in docs]
        return docs, scores

    @abstractmethod
    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieve documents for a given query.

        :param query:
            tokenized query

        :return (docs, scores):
            docs: list of Documents for each batch example.
            scores: [bsz, n_docs] document scores
        """

    def tokenize_query(self, query: str) -> List[int]:
        """
        Tokenize the query.

        :param query:
            query to tokenize

        :return tokenized_query:
            return list of tokens
        """
        return self._tokenizer.encode(query)

    def vectorize_texts(
        self,
        input_text: List[str],
        tokenizer: RagRetrieverTokenizer,
        max_len: Optional[int] = None,
    ) -> torch.LongTensor:
        """
        Vectorize a set of input texts with an arbitrary RagRetrieverTokenizer.

        :param input_text:
            list of input strings
        :param tokenizer:
            tokenizer that encodes the input strings
        :param max_len:
            max length to tokenize

        :return vecs:
            returns a stacked padded tensor of tokens.
        """
        vecs = [tokenizer.encode(q) for q in input_text]
        if max_len:
            vecs = [v[:max_len] for v in vecs]
        vecs, _ = padded_tensor(
            vecs,
            fp16friendly=self.fp16,
            pad_idx=tokenizer.get_pad_idx(),
            max_len=max_len,
        )
        return vecs

    def get_delimiter(self) -> str:
        """
        Return the tokenizer's delimiter.
        """
        return self._tokenizer.get_delimiter()

    def display_docs(self, top_docs: List[List[Document]]):
        """
        Prints documents.

        :param top_docs:
            list of documents for each batch item
        """
        for docs in top_docs:
            for rank, doc in enumerate(docs):
                print(f"Rank: {rank}\n{doc}")

    def share(self) -> TShared:
        """
        Share retriever stuff.

        Share anything that can be handily used by other retrievers.

        This is primarily to share things that take up substantial RAM
        (indices, passages)
        """
        return {}


class RagRetrieverReranker(RagRetriever, ABC):
    """
    Trait that carries methods for Reranker-based retrievers.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        super().__init__(opt, dictionary, shared=shared)
        self.n_final_docs = opt['n_docs']

    @final
    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Perform two-stage retrieval; rescore initial set of docs.

        :param query:
            query tokens

        :return (docs, scores):
            docs: list of Documents for each batch example
            scores: doc scores
        """
        # 1. Get Initial documents
        initial_docs, initial_scores = self._retrieve_initial(query)
        new_scores = self._rescore(query, initial_docs)

        # 2. Get new scores
        final_docs: List[List[Document]] = []
        final_scores: List[torch.Tensor] = []
        new_score_lambda = self._get_new_score_lambda()

        for i in range(len(initial_docs)):
            docs_i = initial_docs[i]
            initial_scores_i = initial_scores[i]
            scores_i = torch.mul(initial_scores_i, (1 - new_score_lambda)) + torch.mul(
                new_scores[i], new_score_lambda
            )
            docs_i, scores_i = argsort_scores_and_docs(
                scores_i, docs_i, self.n_final_docs
            )
            final_docs.append(docs_i)
            final_scores.append(scores_i)

        return final_docs, torch.stack(final_scores)

    @abstractmethod
    def _retrieve_initial(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Perform initial stage of retrieval.

        :param query:
            tokenized query

        :return (docs, scores):
            docs: list of Documents for each batch example
            scores: doc scores
        """

    @abstractmethod
    def _rescore(
        self, query: torch.LongTensor, docs: List[List[Document]]
    ) -> torch.Tensor:
        """
        Rescore retrieved documents.

        :param query:
            tokenized query
        :param docs:
            List of initially retrieved top docs for each batch example

        :return scores:
            return new doc scores.
        """

    @abstractmethod
    def _get_new_score_lambda(self) -> torch.nn.Parameter:
        """
        Return the lambda used for computing the new score.
        """


class DPRRetriever(RagRetriever):
    """
    DPR Retriever.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared=None):
        """
        Initialize DPR Retriever.
        """
        super().__init__(opt, dictionary, shared=shared)
        self.load_index(opt, shared)
        self.n_docs = opt['n_docs']
        self.query_encoder = DprQueryEncoder(
            opt, dpr_model=opt['query_model'], pretrained_path=opt['dpr_model_file']
        )

    def load_index(self, opt, shared):
        if not shared:
            self.indexer = indexer_factory(opt)
            index_path = modelzoo_path(opt['datapath'], opt['path_to_index'])
            passages_path = modelzoo_path(opt['datapath'], opt['path_to_dpr_passages'])
            embeddings_path = None
            if opt['path_to_dense_embeddings'] is not None:
                embeddings_path = modelzoo_path(
                    opt['datapath'], opt['path_to_dense_embeddings']
                )
            self.indexer.deserialize_from(index_path, embeddings_path)
            self.passages = load_passages_dict(passages_path)
        elif shared:
            self.indexer = shared['indexer']
            self.passages = shared['passages']

    def share(self) -> TShared:
        """
        Share FAISS retriever and passages.
        """
        shared = super().share()
        shared['indexer'] = self.indexer
        shared['passages'] = self.passages
        return shared

    def index_retrieve(
        self, query: torch.Tensor, n_docs: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Retrieve over FAISS index.

        :param query:
            bsz x embed_dim query tensor
        :param n_docs:
            number of docs to retrieve

        :return (ids, scores):
            ids: [bsz, n_docs] tensor of document IDs
            scores: [bsz, n_docs] tensor of document scores
        """
        # retrieve docs and scores, reconstruct document embeddings & scores
        # NOTE: important that detach occurs _for retrieval only_, as we use the
        # query encodings to compute scores later in this function; if detached,
        # gradient will not flow to the query encoder.
        top_docs_and_scores = self.indexer.search(
            query.cpu().detach().to(torch.float32).numpy(), n_docs
        )
        ids, np_vectors = zip(*top_docs_and_scores)
        vectors = torch.tensor(np_vectors).to(query)
        if isinstance(self.indexer, DenseHNSWFlatIndexer):
            vectors = vectors[:, :, :-1]
        # recompute exact FAISS scores
        scores = torch.bmm(query.unsqueeze(1), vectors.transpose(1, 2)).squeeze(1)
        if torch.isnan(scores).sum().item():
            raise RuntimeError(
                '\n[ Document scores are NaN; please look into the built index. ]\n'
                '[ This generally happens if FAISS cannot separate vectors appropriately. ]\n'
                '[ If using a compressed index, try building an exact index: ]\n'
                '[ $ python index_dense_embeddings --indexer-type exact... ]'
            )
        ids = torch.tensor([[int(s) for s in ss] for ss in ids])

        return ids, scores

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieve and score.

        For DPR, we encode query tokens and retrieve from FAISS index.

        :param query:
            query tokens

        :return (docs, scores):
            docs: list of (text, title) tuples for each batch example
            scores: doc scores
        """
        query_enc = self.query_encoder(query)
        top_doc_ids_tensor, top_doc_scores = self.index_retrieve(query_enc, self.n_docs)
        top_docs, top_doc_ids = [], []
        for i in range(query.size(0)):
            ids_i = []
            docs_i = []
            for int_id in top_doc_ids_tensor[i]:
                doc_id = str(int_id.item())
                passage = self.passages[doc_id]

                ids_i.append(doc_id)
                docs_i.append(Document(title=passage[1], text=passage[0], docid=doc_id))
            top_docs.append(docs_i)
            top_doc_ids.append(ids_i)
        return top_docs, top_doc_scores


class TFIDFRetriever(RagRetriever):
    """
    Use TFIDF to retrieve wikipedia documents.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        """
        Init a TFIDFRetrieverAgent.
        """
        opt['query_model'] = 'tfidf'
        super().__init__(opt, dictionary, shared=shared)
        tfidf_opt = {
            'model': 'rag_tfidf_retriever',
            'model_file': (opt['tfidf_model_path']),
            'tfidf_model_path': opt['tfidf_model_path'],
            'retriever_num_retrieved': opt['n_docs'],
            'retriever_mode': 'keys',
            'override': {'model': 'rag_tfidf_retriever', 'remove_title': False},
        }
        self.n_docs = opt['n_docs']
        self.max_doc_paragraphs = opt['tfidf_max_doc_paragraphs']
        assert self.max_doc_paragraphs != 0
        if not shared:
            self.tfidf_retriever = create_agent(tfidf_opt)
        else:
            self.tfidf_retriever = shared['tfidf_retriever']

    def share(self) -> TShared:
        shared = super().share()
        shared['tfidf_retriever'] = self.tfidf_retriever
        return shared

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], Union[torch.Tensor, List[torch.Tensor]]]:
        """
        Retrieve and score using TFIDF.

        :param query:
            query tokens

        :return (docs, scores):
            docs: list of (text, title) tuples for each batch example
            scores: doc scores
        """

        def _build_doc(idx, cand):
            title = cand.split('\n\n')[0]
            paragraphs = cand.split('\n\n')[1:]
            if self.max_doc_paragraphs > 0:
                paragraphs = paragraphs[: self.max_doc_paragraphs]
            return Document(title=title, text=' '.join(paragraphs), docid=ids_i[idx])

        docs = []
        scores = []

        for q in query:
            query_text = self._tokenizer.decode(q)
            self.tfidf_retriever.observe({'text': query_text, 'episode_done': True})
            act = self.tfidf_retriever.act()
            if 'candidate_scores' not in act:
                scores_i = [0] * self.n_docs
                docs_i = [BLANK_DOC] * self.n_docs
            else:
                scores_i = act['candidate_scores']
                candidate_docs = act['text_candidates']
                ids_i = act['candidate_ids']
                candidate_docs = [
                    _build_doc(j, c) for j, c in enumerate(act['text_candidates'])
                ]
                docs_i = candidate_docs[: self.n_docs]
                scores_i = scores_i[: self.n_docs]
                if len(docs_i) < self.n_docs:
                    # Something went wrong with TFIDF here; need to add null docs
                    logging.warning(
                        f'Ex has less than {self.n_docs} TFIDF docs: {len(docs_i)}'
                    )
                    num_null = self.n_docs - len(docs_i)
                    docs_i += [BLANK_DOC] * num_null
                    scores_i = np.append(scores_i, [0] * num_null)
            docs.append(docs_i)
            scores.append(torch.FloatTensor(scores_i).to(query.device))

        scores = torch.stack(scores)
        return docs, scores


class DPRThenTorchReranker(RagRetrieverReranker, DPRRetriever, ABC):
    """
    Base Class for DPR --> TorchRanker Retrievers.

    Handles some shared functionality.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        """
        Initialize DPR model.

        It is up to subclasses to initialize rerankers.
        """
        RagRetrieverReranker.__init__(self, opt, dictionary, shared=shared)
        self.dpr_num_docs = opt['dpr_num_docs']
        assert self.dpr_num_docs
        dpr_opt = copy.deepcopy(opt)
        dpr_opt['n_docs'] = self.dpr_num_docs
        DPRRetriever.__init__(self, dpr_opt, dictionary, shared=shared)

    def get_reranker_opts(self, opt: Opt) -> Dict[str, Any]:
        """
        Provide options used when building the rerankers.

        Base class ensures that various optimizations (cuda, fp16, parallel)
        are accounted for.

        :param opt:
            base opt

        :return options_dict:
            return a dictionary mapping options to values.
        """
        return {
            'no_cuda': opt['no_cuda'],
            'fp16': opt['fp16'],
            'model_parallel': opt['model_parallel'],
            'data_parallel': opt['data_parallel'],
        }

    def _build_reranker(
        self, opt: Opt
    ) -> Tuple[torch.nn.Module, RagRetrieverTokenizer]:
        """
        Builds reranker.

        :param opt:
            original opt

        :return (module, dict)
            module: the model from the agent created via the options
            dict: A RagRetrieverTokenizer, dictionary for the created model.
        """
        rerank_opt = copy.deepcopy(opt)
        rerank_opt = {**TRANSFORMER_RANKER_BASE_OPT, **self.get_reranker_opts(opt)}
        logging.disable()
        agent = create_agent(rerank_opt)
        logging.enable()
        assert isinstance(agent, TorchRankerAgent)

        return agent.model, RagRetrieverTokenizer('', agent.dict, max_length=360)

    def _retrieve_initial(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Initial DPR retrieval.

        Just call superclass to retrieve first stage.

        :param query:
            encoding of query
        :param mask:
            optional query mask

        :return (docs, scores):
            docs: list of (text, title) tuples for each batch example
            scores: doc scores
        """
        return DPRRetriever.retrieve_and_score(self, query)


class DPRThenPolyRetriever(DPRThenTorchReranker):
    """
    2 Stage Retrieval with DPR and Poly-encoder.

    1. Retrieve N Docs with DPR
    2. Rescore docs with polyencoder
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        """
        Initialize a Poly-Encoder Agent.
        """
        # 1. Call super to init DPR
        super().__init__(opt, dictionary, shared=shared)

        # 2. Poly-encoder
        self.polyencoder, self.poly_tokenizer = self._build_reranker(opt)
        self.register_parameter(
            "poly_score_lambda",
            torch.nn.Parameter(torch.tensor([float(opt['poly_score_initial_lambda'])])),
        )

    def _get_new_score_lambda(self) -> torch.nn.Parameter:
        """
        Return the lambda used for computing the new score.
        """
        return self.poly_score_lambda  # type: ignore

    def get_reranker_opts(self, opt: Opt) -> Dict[str, Any]:
        """
        Provide options used when building the polyencoder.

        :param opt:
            base opt

        :return options_dict:
            return a dictionary mapping options to values.
        """
        from parlai.agents.rag.args import PRETRAINED_RANKER_TYPES

        init_path = opt['polyencoder_init_model']
        if init_path in PRETRAINED_RANKER_TYPES:
            init_model = f"zoo:pretrained_transformers/poly_model_huge_{opt['polyencoder_init_model']}/model"
            dict_file = f"zoo:pretrained_transformers/poly_model_huge_{opt['polyencoder_init_model']}/model.dict"
        else:
            assert os.path.exists(init_path)
            init_model = init_path
            dict_file = f'{init_path}.dict'

        return {
            'model': 'transformer/polyencoder',
            'init_model': init_model,
            'dict_file': dict_file,
            # necessary opt args
            'multitask_weights': [1],
            **{k: opt[k] for k in POLYENCODER_OPT_KEYS},
            **super().get_reranker_opts(opt),
        }

    def _rescore(
        self, query: torch.LongTensor, docs: List[List[Document]]
    ) -> torch.Tensor:
        """
        Compute Poly-encoder score with initial set of Documents.

        Scoring taken from PolyencoderAgent.score_candidates

        :param query:
            query tokens, used in DPR retrieval.
        :param docs:
            List of initially retrieved top docs for each batch example

        :return new_scores:
            return scored documents.
        """
        poly_query_vec = self.vectorize_texts(
            [self._tokenizer.decode(q) for q in query],
            self.poly_tokenizer,
            self.max_query_len,
        ).to(query.device)

        doc_vecs = torch.stack(
            [
                self.vectorize_texts(
                    [d.get_tokenization_str() for d in docs_i],
                    self.poly_tokenizer,
                    self.max_doc_len,
                )
                for docs_i in docs
            ]
        ).to(query.device)

        ctxt_rep, ctxt_rep_mask, _ = self.polyencoder(ctxt_tokens=poly_query_vec)
        _, _, cand_rep = self.polyencoder(cand_tokens=doc_vecs)
        scores = self.polyencoder(
            ctxt_rep=ctxt_rep, ctxt_rep_mask=ctxt_rep_mask, cand_rep=cand_rep
        )
        return scores


class PolyFaissRetriever(DPRThenPolyRetriever):
    """
    Poly-encoder Retriever, using FAISS.

    Performs FAISS retrieval to retrieve N initial docs; re-ranks according to Poly-
    encoder score to narrow down to K docs.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared = None):
        assert opt['query_model'] == 'dropout_poly'
        super().__init__(opt, dictionary, shared=shared)
        self.dropout_poly = RagDropoutPolyWrapper(opt)
        self.polyencoder = self.dropout_poly.model

        self.poly_tokenizer = RagRetrieverTokenizer(
            opt['query_model'], self.dropout_poly.dict, max_length=360
        )

        model = (
            self.polyencoder.module
            if hasattr(self.polyencoder, 'module')
            else self.polyencoder
        )
        for param in model.encoder_cand.parameters():  # type: ignore
            # freeze document encoding for PolyFAISS.
            param.requires_grad = False


@register_agent("rag_tfidf_retriever")
class RagTfidfRetrieverAgent(TfidfRetrieverAgent):
    """
    Wrapper around TFIDF Retriever to cache retrieved documents.
    """

    def __init__(self, opt: Opt, shared: TShared = None):
        super().__init__(opt, shared)
        if not shared:
            self.docid_to_text = {}
        else:
            self.docid_to_text = shared.get('docid_to_text', {})

    def share(self) -> TShared:
        shared = super().share()
        shared['docid_to_text'] = self.docid_to_text
        return shared

    def doc2txt(self, docid):
        """
        Cache document texts during train/eval.
        """
        if docid not in self.docid_to_text:
            text = super().doc2txt(docid)
            self.docid_to_text[docid] = text
        else:
            text = self.docid_to_text[docid]
        return text


BLANK_SEARCH_DOC = {'url': None, 'content': '', 'title': ''}
NO_SEARCH_QUERY = 'no_passages_used'


class SearchQueryRetriever(RagRetriever):
    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared):
        RagRetriever.__init__(self, opt, dictionary, shared=shared)
        opt['skip_retrieval_token'] = NO_SEARCH_QUERY
        self.n_docs = opt['n_docs']
        self.len_chunk = opt['splitted_chunk_length']
        self.doc_chunk_split_mode = opt['doc_chunk_split_mode']
        n_doc_chunks = opt['n_ranked_doc_chunks']
        chunk_ranker_type = opt['doc_chunks_ranker']
        if chunk_ranker_type == 'tfidf':
            self.chunk_reranker = TfidfChunkRanker(n_doc_chunks)
        else:
            self.chunk_reranker = HeadChunkRanker(n_doc_chunks)

        if not shared:
            self.query_generator = self.init_search_query_generator(opt)
        else:
            self.query_generator = shared['query_generator']
        self.dict = dictionary
        self.init_query_encoder(opt)

    def share(self) -> TShared:
        shared = super().share()
        shared['query_generator'] = self.query_generator
        return shared

    def init_search_query_generator(self, opt) -> TorchGeneratorAgent:
        model_file = opt['search_query_generator_model_file']
        logging.info('Loading search generator model')
        logging.disable()
        search_query_gen_agent = create_agent_from_model_file(
            model_file,
            opt_overrides={
                'skip_generation': False,
                'inference': opt['search_query_generator_inference'],
                'beam_min_length': opt['search_query_generator_beam_min_length'],
                'beam_size': opt['search_query_generator_beam_size'],
                'text_truncate': opt['search_query_generator_text_truncate'],
            },
        )
        logging.enable()
        logging.info('Search query generator model loading completed!')
        return search_query_gen_agent

    def generate_search_query(self, query: torch.LongTensor) -> List[str]:
        """
        Generates a list of queries for the encoded query (context) tensor.
        """
        texts = [self._tokenizer.decode(q) for q in query]
        obs_list = []
        for t in texts:
            msg = Message({'text': t, 'episode_done': True})
            obs_list.append(self.query_generator.observe(msg))
            self.query_generator.reset()  # Erase the history
        search_quries = [r['text'] for r in self.query_generator.batch_act(obs_list)]
        logging.debug(f'Generated search queries {search_quries}')
        return search_quries

    def init_query_encoder(self, opt):
        if hasattr(self, 'query_encoder'):
            # It is already instantiated
            return
        self.query_encoder = DprQueryEncoder(
            opt, dpr_model=opt['query_model'], pretrained_path=opt['dpr_model_file']
        )

    def text2tokens(self, txt: str):
        if self.doc_chunk_split_mode == 'word':
            return txt.split(' ')
        else:
            return self.dict.txt2vec(txt)

    def tokens2text(self, tokens: List[int]):
        if self.doc_chunk_split_mode == 'word':
            return ' '.join(tokens)
        else:
            return self.dict.vec2txt(tokens)

    def pick_chunk(self, query, doc_text):
        """
        Splits the document and returns the selected chunks.

        The number of returned chunks is controlled by `n_ranked_doc_chunks` in opt. The
        chunk selection is determined by `doc_chunks_ranker` in the opt.
        """
        if not doc_text:
            # When there is no search query for the context
            return [("", 0)]
        tokens = self.text2tokens(doc_text)
        doc_chunks = [
            self.tokens2text(tokens[i : i + self.len_chunk])
            for i in range(0, len(tokens), self.len_chunk)
        ]
        return self.chunk_reranker.get_top_chunks(query, doc_chunks)


class SearchQuerySearchEngineRetriever(SearchQueryRetriever):
    """
    A retriever that uses a search engine server for rertrieving documents.

    It instantiates a `SearchEngineRetriever` object that in turns send search queries
    to an external server for retrieving documents.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared: TShared):
        super().__init__(opt, dictionary, shared)
        if not shared:
            self.search_client = self.initiate_retriever_api(opt)
        else:
            self.search_client = shared['search_client']

    def share(self) -> TShared:
        shared = super().share()
        shared['search_client'] = self.search_client
        return shared

    def initiate_retriever_api(self, opt) -> SearchEngineRetriever:
        logging.info('Creating the search engine retriever.')
        return SearchEngineRetriever(opt)

    def _empty_docs(self, num: int):
        """
        Generates the requested number of empty documents.
        """
        return [BLANK_SEARCH_DOC for _ in range(num)]

    def rank_score(self, rank_id: int):
        """
        Scores the chunks of the retrieved document based on their rank.

        Note that this is the score for the retrieved document and applies to all its
        chunks.
        """
        return 1 / (1 + rank_id)

    def _display_urls(self, search_results):
        """
        Generates a string that lists retrieved URLs (document IDs).
        """
        return '\n'.join([d['url'] for d in search_results if d['url']])

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieves relevant documents for the query (the conversation context).

        This method conducts three main steps that are flagged in the main code as well.
        step 1: generate search queries for the conversation context batch.     This
        step uses the query generator model (self.query_generator). step 2: use the
        search client to retrieve documents.     This step uses retrieval API agent
        (self.search_client) step 3: generate the list of Document objects from the
        retrieved content.     Here if the documents too long, the code splits them and
        chooses a chunk     based on the selected `doc_chunks_ranker` in the opt.
        """
        # step 1
        search_queries = self.generate_search_query(query)

        # step 2
        search_results_batach = self.search_client.retrieve(search_queries, self.n_docs)

        # step 3
        top_docs = []
        top_doc_scores = []
        for sq, search_results in zip(search_queries, search_results_batach):
            if not search_results:
                search_results = self._empty_docs(self.n_docs)
            elif len(search_results) < self.n_docs:
                remain_docs = self.n_docs - len(search_results)
                search_results.extend(self._empty_docs(remain_docs))
            docs_i = []
            scors_i = []
            # Change this debug later
            logging.debug(f'URLS:\n{self._display_urls(search_results)}')
            for i, doc in enumerate(search_results):
                url = doc['url']
                title = doc['title']
                dcontent = doc['content']
                assert type(dcontent) in (
                    str,
                    list,
                ), f'Unrecognized retrieved doc: {dcontent}'
                full_text = (
                    dcontent if isinstance(dcontent, str) else '\n'.join(doc['content'])
                )
                doc_chunks = [dc[0] for dc in self.pick_chunk(sq, full_text)]
                for splt_id, splt_content in enumerate(doc_chunks):
                    docs_i.append(
                        Document(
                            docid=url, text=splt_content, title=f'{title}_{splt_id}'
                        )
                    )
                    scors_i.append(self.rank_score(i))
            top_docs.append(docs_i)
            top_doc_scores.append(scors_i)
        self.top_docs = top_docs
        return top_docs, torch.Tensor(top_doc_scores).to(query.device)


class SearchQueryFAISSIndexRetriever(SearchQueryRetriever, DPRRetriever):
    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared):
        SearchQueryRetriever.__init__(self, opt, dictionary, shared=shared)
        self.load_index(opt, shared)

    def share(self) -> TShared:
        shared = SearchQueryRetriever.share(self)
        shared.update(DPRRetriever.share(self))
        return shared

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Retrieves from the FAISS index using a search query.

        This methods relies on the `retrieve_and_score` method in `RagRetriever`
        ancestor class. It receive the query (conversation context) and generatess the
        search term queries based on them. Then uses those search quries (instead of the
        the query text itself) to retrieve from the FAISS index.
        """

        search_queries = self.generate_search_query(query)
        tokenized_search_queries, _ = padded_tensor(
            [self._tokenizer.encode(sq) for sq in search_queries]
        )
        top_docs, top_doc_scores = DPRRetriever.retrieve_and_score(
            self, tokenized_search_queries.to(query.device)
        )
        for query_id in range(len(top_docs)):
            if search_queries[query_id] == NO_SEARCH_QUERY:
                top_docs[query_id] = [BLANK_DOC for _ in range(self.n_docs)]
        return top_docs, top_doc_scores


class DocumentChunkRanker:
    """
    Base class for controlling splitting long documents and selecting relevant chunks.
    """

    def __init__(self, n_retrieved_chunks):
        self.n_ret_chunks = n_retrieved_chunks

    @abstractmethod
    def get_top_chunks(self, query, doc_chunks):
        """
        Ranks documents (chunk) based on their relevance to `query`
        """


class HeadChunkRanker(DocumentChunkRanker):
    """
    Returns the head chunks only.
    """

    def get_top_chunks(self, query, doc_chunks):
        """
        Return chunks in doc-present order.
        """
        return [(c,) for c in doc_chunks[: self.n_ret_chunks]]


class TfidfChunkRanker(DocumentChunkRanker):
    """
    Uses TF-IDF to compare chunks to the original search query.
    """

    def __init__(self, n_retrieved_chunks):
        super().__init__(n_retrieved_chunks)
        self._vectorizer = TfidfVectorizer()

    def get_top_chunks(self, query, doc_chunks):
        vectorized_corpus = self._vectorizer.fit_transform(doc_chunks + [query])
        docs_vec = vectorized_corpus[:-1, :]
        q_vec = vectorized_corpus[-1, :]
        scores = np.hstack((q_vec * docs_vec.transpose()).toarray())
        top_chunk_ids = np.argsort(-scores)[: self.n_ret_chunks]
        return [(doc_chunks[i], scores[i]) for i in top_chunk_ids]


def retriever_factory(
    opt: Opt, dictionary: DictionaryAgent, shared=None
) -> Optional[RagRetriever]:
    """
    Build retriever.

    :param opt:
        ParlAI Opt
    :param dictionary:
        dictionary agent
    :param shared:
        shared objects.

    :return retriever:
        return a retriever for RAG.
    """
    if opt.get('converting'):
        return None
    # only build retriever when not converting a BART model
    retriever = RetrieverType(opt['rag_retriever_type'])
    if retriever is RetrieverType.DPR:
        return DPRRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.TFIDF:
        return TFIDFRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.DPR_THEN_POLY:
        return DPRThenPolyRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.POLY_FAISS:
        return PolyFaissRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.SEARCH_ENGINE:
        return SearchQuerySearchEngineRetriever(opt, dictionary, shared=shared)
    elif retriever is RetrieverType.SEARCH_TERM_FAISS:
        return SearchQueryFAISSIndexRetriever(opt, dictionary, shared=shared)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Dropout Poly-encoder.

The dropout poly-encoder mimics a combination of a bi-encoder & poly-encoder during
training, and during inference is exactly a poly-encoder.
"""
from parlai.core.params import ParlaiParser
from parlai.agents.transformer.polyencoder import PolyencoderAgent
from parlai.core.opt import Opt
from parlai.core.torch_agent import Batch

import random
import torch
from typing import Optional, Tuple


def reduce_ctxt(
    ctxt: torch.Tensor, mask: torch.Tensor, reduction_type: str
) -> torch.Tensor:
    """
    Reduce ctxt tensor to appropriate dimension for bi-encoder scoring.

    Functionality copied from TransformerEncoder.reduce_output.

    :param ctxt:
        all transformer outputs.
    :param mask:
        context mask
    :param reduction_type:
        how to reduce the context

    :return reduced:
        return reduced context tensor
    """
    if reduction_type == 'first':
        return ctxt[:, 0, :]
    elif reduction_type == 'max':
        return ctxt.max(dim=1)[0]
    elif reduction_type == 'mean':
        divisor = mask.float().sum(dim=1).unsqueeze(-1).clamp(min=1).type_as(ctxt)
        output = ctxt.sum(dim=1) / divisor
        return output
    else:
        raise ValueError("Can't handle --reduction-type {}".format(reduction_type))


class DropoutPolyAgent(PolyencoderAgent):
    """
    Dropout Polyencoder Agent.

    Overrides score_candidates to, sometimes, score candidates in the bi-encoder method.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('Dropout Polyencoder Arguments')
        agent.add_argument(
            '--poly-dropout-prob',
            type=float,
            default=0.5,
            help='How often to switch to a bi-encoder setup during training',
        )
        agent.add_argument(
            '--poly-dropout-reduction-type',
            type=str,
            choices=['first', 'max', 'mean'],
            default='first',
            help='how to reduce output when ignoring poly-codes during training',
        )
        agent.add_argument(
            '--poly-dropout-use-codes',
            type='bool',
            default=True,
            help='Whether to attend over codes prior to dropout method.',
        )

        return parser

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.poly_dropout_reduction_type = opt['poly_dropout_reduction_type']
        self.poly_dropout_prob = opt['poly_dropout_prob']
        self.use_codes = opt.get('poly_dropout_use_codes', True)
        assert 0 <= self.poly_dropout_prob <= 1

    def get_ctxt_rep(self, batch: Batch) -> Tuple[torch.Tensor, torch.BoolTensor]:
        """
        Encode context representation.
        """
        if self.use_codes:
            ctxt_rep, ctxt_rep_mask, _ = self.model(**self._model_context_input(batch))
        else:
            # In dataparallel, the model is the `model.module`
            model = self.model.module if hasattr(self.model, 'module') else self.model
            model.type = 'n_first'
            ctxt_rep, ctxt_rep_mask, _ = self.model(**self._model_context_input(batch))
            model.type = self.opt['polyencoder_type']

        return ctxt_rep, ctxt_rep_mask

    def get_scores(
        self,
        ctxt_rep: torch.Tensor,
        ctxt_rep_mask: torch.BoolTensor,
        cand_rep: torch.Tensor,
    ) -> torch.Tensor:
        """
        Score context with candidates.

        During training, with some probability we fall back to the Bi-encoder scoring
        method.
        """
        if self.is_training and random.random() < self.poly_dropout_prob:
            ctxt_rep = reduce_ctxt(
                ctxt_rep, ctxt_rep_mask, self.poly_dropout_reduction_type
            )
            scores = torch.bmm(ctxt_rep.unsqueeze(1), cand_rep.transpose(1, 2)).squeeze(
                1
            )
        else:
            scores = self.model(
                ctxt_rep=ctxt_rep, ctxt_rep_mask=ctxt_rep_mask, cand_rep=cand_rep
            )
        return scores



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://download.microsoft.com/download/E/B/8/EB84CB1A-D57D-455F-B905-3ABDE80404E5/metalwoz-v1.zip',
        'metalwoz-v1.zip',
        '2a2ae3b25760aa2725e70bc6480562fa5d720c9689a508d28417631496d6764f',
    ),
    DownloadableFile(
        'https://download.microsoft.com/download/0/c/4/0c4a8893-cbf9-4a43-a44a-09bab9539234/metalwoz-test-v1.zip',
        'metalwoz-test-v1.zip',
        '6722d1d9ec05334dd801972767ae3bdefcd15f71bf73fea1d098f214a96a7c6c',
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'metalwoz')
    version = '1.0'

    if not build_data.built(dpath, version_string=version):
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)
        build_data.make_dir(os.path.join(dpath, 'train', 'dialogues'))
        build_data.make_dir(os.path.join(dpath, 'test', 'dialogues'))

        # Download the data.
        RESOURCES[0].download_file(os.path.join(dpath, 'train'))
        RESOURCES[1].download_file(os.path.join(dpath, 'test'))

        build_data.untar(os.path.join(dpath, 'test'), 'dstc8_metalwoz_heldout.zip')
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Tools for annotating modules with lightweight dependency injection.

Primarily designed to swap out individual modules deep within the transformer class hierarchy.

Usage:

```python
@swappable(component=DefaultClass, ...)
class SomeModel(nn.Module):
    ...
```

Within the model, access the swappable classes like so:

```python
self.swappables.component()
```

When instantiating the model, swap out the component like so:

```python
model = SomeModel.with_components(component=NewCustomClass)()
```
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar


C = TypeVar('C', bound=object)


class ModularComponent(Generic[C]):
    @dataclass
    class SwappableSubcomponents:
        """
        Define any swappable subcomponents by adding the class (or a constructor) of the
        components as attributes of this object.

        When using the @swappable decorator, this class is created programmatically.
        """

    @classmethod
    def with_components(cls, **kwargs) -> ModularComponentBuilder[ModularComponent[C]]:
        return ModularComponentBuilder(
            klass=cls, subcomponents=cls.SwappableSubcomponents(**kwargs)
        )

    def __init__(
        self, *args, subcomponents: Optional[SwappableSubcomponents] = None, **kwargs
    ):
        """
        Unpacks the swappable_components, then forwards the call up the MRO chain.
        """
        self.swappables = subcomponents or type(self).SwappableSubcomponents()
        assert (
            type(self.swappables) is not ModularComponent.SwappableSubcomponents
        ), "Modular components must declare their own SwappableSubcomponents"
        super().__init__(*args, **kwargs)


MC = TypeVar('MC', bound=ModularComponent)


class ModularComponentBuilder(Generic[MC]):
    """
    When a component has swappable subcomponents, use this object to specify both the
    component type and it's subcomponent types at the same time.
    """

    def __init__(
        self,
        klass: Type[MC],
        subcomponents: Optional[ModularComponent.SwappableSubcomponents] = None,
    ) -> None:
        self._klass = klass
        self._subcomponents = subcomponents or klass.SwappableSubcomponents()

    def __call__(self, *args, **kwargs) -> MC:
        """
        Forward calls to this instance to __init__ of wrapped class.
        """
        return self._klass(*args, subcomponents=self._subcomponents, **kwargs)

    @property
    def swappables(self) -> Any:
        return self._subcomponents

    def swap_components(self, **kwargs) -> None:
        for name, value in kwargs.items():
            setattr(self._subcomponents, name, value)


def swappable(**kwargs) -> Callable[[Type[C]], Type[C]]:
    """
    Decorator that adds swappable subcomponents to a class.

    Usage:

    ```python
    @swappable(component_name=DefaultComponentClass, ...)
    ```
    """

    # Decorators need to return callables that accept only the decorated object.
    # To comply, bundle kwargs into a function that accepts only one argument.
    def wrap(cls: Type[C]) -> Type[C]:
        return _make_class_swappable(cls, **kwargs)

    return wrap


def _make_class_swappable(cls: Type[C], **kwargs) -> Type[C]:
    """
    Creates a new class that subclasses ModularComponent.

    Modifies that class to to accept constructors for the components passed to the
    decorator.
    """

    def _class_dict(new_fields) -> Dict[str, Any]:
        """
        Sets up the class attributes, along with type annotations.
        """
        return {
            **new_fields,
            '__annotations__': {**{k: type(v) for k, v in new_fields.items()}},
        }

    # Create SwappableSubcomponents dataclass with components passed to @swappable
    subcomponent_class_name = ModularComponent.SwappableSubcomponents.__name__
    subcomponent_dataclass = dataclass(
        type(subcomponent_class_name, (), _class_dict(kwargs))
    )

    # Create a new class that subclasses the decorated class. This new class holds
    # the SwappableSubcomponents dataclass created above (list of components that are
    # swappable) and the swappables attribute (the actual swappable component constructors).
    return type(
        # We append "_Swappable" to the new class name for transparency.
        f"{cls.__name__}_Swappable",
        # ModularComponent comes before the class so we can intercept __init__ calls.
        (ModularComponent, cls),  # type: ignore
        # Items in this dictionary are converted to class attributes by type()
        _class_dict({subcomponent_class_name: subcomponent_dataclass}),
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
FAISS-based Indexers.

Adapted from https://github.com/facebookresearch/DPR/blob/master/dpr/indexer/faiss_indexers.py
"""
from parlai.core.build_data import modelzoo_path
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from abc import ABC, abstractmethod
import math
import numpy as np
import os
import random
import time
import torch
import torch.cuda
from typing import List, Tuple, Optional

from parlai.agents.rag.args import WIKIPEDIA_COMPRESSED_INDEX, WIKIPEDIA_EXACT_INDEX


def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks.

    From https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    from itertools import zip_longest

    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


class BaseIndexer(ABC):
    """
    Top-Level Indexer.

    Provides an interface for interacting with FAISS Indexes.
    """

    def __init__(self, opt: Opt):
        self.buffer_size = opt['indexer_buffer_size']
        self.index_id_to_db_id = []
        self.index = None
        try:
            import faiss  # noqa: f401

            self.faiss = faiss
        except ImportError:
            raise ImportError(
                'Please install faiss: https://github.com/facebookresearch/faiss/blob/master/INSTALL.md'
            )

    @abstractmethod
    def index_data(self, data: List[torch.Tensor]):
        """
        Index data.

        Given a list of tensors, construct a FAISS Index.

        :param data:
            list of torch.Tensors
        """

    def get_search_vectors(self, query_vectors: np.array) -> np.array:
        """
        Given array of query vectors, return the vectors used for search.

        Allows subclasses to modify the vectors at their discretion.

        :param query_vectors:
            q vecs to modify

        :return search vectors:
            return vectors used for FAISS.
        """
        return query_vectors

    def search(
        self, query_vectors: np.array, top_docs: int
    ) -> List[Tuple[List[int], List[np.array]]]:
        """
        Search FAISS index.

        :param query_vectors:
            query vectors into the index
        :param top_docs:
            number of docs to return

        :return top_docs:
            returns, for each query vector:
                a list of document ids (according to db),
                a list of reconstructed document vectors
        """
        query_vectors = self.get_search_vectors(query_vectors)
        logging.debug(f'query_vectors {query_vectors.shape}')
        _scores, indexes, vectors = self.index.search_and_reconstruct(
            query_vectors, top_docs
        )
        db_ids = [
            [self.index_id_to_db_id[i] for i in query_top_idxs]
            for query_top_idxs in indexes
        ]
        result = [(db_ids[i], vectors[i]) for i in range(len(db_ids))]
        return result

    def serialize(self, file: str):
        """
        Serialize index into file.

        :param file:
            output file.
        """
        logging.info(f'Serializing index to {file}')

        if os.path.isdir(file):
            index_file = os.path.join(file, "index")
            meta_file = os.path.join(file, "index_meta")
        else:
            index_file = f'{file}.index'
            meta_file = f'{file}.index_meta'

        self.faiss.write_index(self.index, index_file)
        if self.index_id_to_db_id:
            torch.save(self.index_id_to_db_id, meta_file)

    def deserialize_from(self, file: str, emb_path: Optional[str] = None):
        """
        Deserialize index from file.

        :param file:
            input file
        :param emb_path:
            optional path to embeddings
        """
        logging.info(f'Loading index from {file}')

        if os.path.isdir(file):
            index_file = os.path.join(file, "index")
            meta_file = os.path.join(file, "index_meta")
        elif not file.endswith('.index'):
            index_file = f'{file}.index'
            meta_file = f'{file}.index_meta'
        else:
            index_file = file
            meta_file = f'{index_file}_meta'

        self.index = self.faiss.read_index(index_file)
        logging.info(f'Loaded index of type {self.index} and size {self.index.ntotal}')

        if os.path.exists(meta_file):
            self.index_id_to_db_id = torch.load(meta_file)
        else:
            index_dir = os.path.split(file)[0] if emb_path is None else emb_path
            if not os.path.isdir(index_dir):
                # if emb_path has the embeddings name in there, need to split.
                index_dir = os.path.split(index_dir)[0]
            meta_files = [f for f in os.listdir(index_dir) if f.startswith('ids_')]
            meta_files = sorted(meta_files, key=lambda x: int(x.split('_')[-1]))
            for f in meta_files:
                ids = torch.load(os.path.join(index_dir, f))
                self.index_id_to_db_id.extend(ids)
            torch.save(self.index_id_to_db_id, meta_file)
        assert (
            len(self.index_id_to_db_id) == self.index.ntotal
        ), 'Deserialized index_id_to_db_id should match faiss index size '
        f'{len(self.index_id_to_db_id)} != {self.index.ntotal}'


class DenseHNSWFlatIndexer(BaseIndexer):
    """
    Indexer for Dense, HNSW FLAT Index.

    High retrieval accuracy comes at the cost of High RAM Usage.
    """

    def __init__(self, opt: Opt):
        super().__init__(opt)
        # IndexHNSWFlat supports L2 similarity only
        # so we have to apply DOT -> L2 similairy space conversion with the help of an extra dimension
        index = self.faiss.IndexHNSWFlat(
            opt['retriever_embedding_size'] + 1, opt['hnsw_indexer_store_n']
        )
        index.hnsw.efSearch = opt['hnsw_ef_search']
        index.hnsw.efConstruction = opt['hnsw_ef_construction']
        self.index = index
        self.built = False

    def index_data(self, tensors: List[torch.Tensor]):
        """
        Index data.

        The HNSW Flat Indexer computes an auxiliary dimension that converts inner product
        similarity to L2 distance similarity.

        :param data:
            List of torch.Tensor
        """
        data = torch.cat(tensors).float()
        n = data.size(0)
        # max norm is required before putting all vectors in the index to convert inner product similarity to L2
        if self.built:
            raise RuntimeError(
                'HNSW index needs to index all data at once, results will be unpredictable otherwise.'
            )
        phi = 0
        norms = (data ** 2).sum(dim=1)
        max_norms = norms.max().item()
        phi = max(phi, max_norms)
        logging.info(f'HNSWF DotProduct -> L2 space phi={phi}')
        start = time.time()

        for i in range(0, n, self.buffer_size):
            vectors_i = data[i : i + self.buffer_size]
            norms_i = norms[i : i + self.buffer_size]
            aux_dims = torch.sqrt(phi - norms_i)
            hnsw_vectors = torch.cat([vectors_i, aux_dims.unsqueeze(1)], dim=1)
            self.index.add(hnsw_vectors.numpy())
            logging.info(
                f'{time.time() - start}s Elapsed: data indexed {i + len(vectors_i)}'
            )

        logging.info(f'Total data indexed {n}')

    def get_search_vectors(self, query_vectors: np.array) -> np.array:
        """
        Add an additional dimension to the query vectors to account for cosine
        similarity.

        :param query_vectors:
            query vectors of dimension [n_search, q_dim]

        :return search_vectors:
            search vectors of dimension [n_search, q_dim + 1]
        """
        aux_dim = np.zeros(len(query_vectors), dtype='float32')
        query_hnsw_vectors = np.hstack((query_vectors, aux_dim.reshape(-1, 1)))
        return query_hnsw_vectors

    def deserialize_from(self, file: str, emb_path: Optional[str] = None):
        super().deserialize_from(file, emb_path)
        # to trigger warning on subsequent indexing
        self.built = True


class CompressedIndexer(BaseIndexer):
    """
    Compressed Indexer.

    If a FAISS index factory is specified, we build that.
    (see https://github.com/facebookresearch/faiss/wiki/The-index-factory)

    The default is IVF4096_HNSW128,PQ128; this also translates directly to the
    default when index_factory is specified as ''.

    See https://github.com/facebookresearch/faiss/wiki/Faiss-indexes#cell-probe-methods-indexivf-indexes
    for more details.
    """

    def __init__(self, opt: Opt):
        """
        Initialize IVFPQ FAISS Indexer.

        The IVFPQ Indexer is a great way to reduce memory footprint of dense embeddings.
        """
        super().__init__(opt)
        self.dim = opt['retriever_embedding_size']
        self.use_gpu_train = (
            not opt['no_cuda']
            and torch.cuda.is_available()
            and opt['compressed_indexer_gpu_train']
        )
        self.hnsw_ef_search = opt['hnsw_ef_search']

        self.index_factory = opt['compressed_indexer_factory']
        if self.index_factory:
            logging.warning(f'Creating Index from Index Factory: {self.index_factory}')
            self.is_ivf_index = 'IVF' in self.index_factory
            self.index = self.faiss.index_factory(
                self.dim, self.index_factory, self.faiss.METRIC_INNER_PRODUCT
            )
        else:
            self.is_ivf_index = True
            quantizer = self.faiss.IndexHNSWFlat(
                self.dim, opt['hnsw_indexer_store_n'], self.faiss.METRIC_INNER_PRODUCT
            )
            quantizer.hnsw.efConstruction = opt['hnsw_ef_construction']
            quantizer.hnsw.efSearch = opt['hnsw_ef_search']
            ivf_index = self.faiss.IndexIVFPQ(
                quantizer, self.dim, 4096, 128, 8, self.faiss.METRIC_INNER_PRODUCT
            )
            ivf_index.nprobe = opt['compressed_indexer_nprobe']
            self.index = ivf_index

        if self.is_ivf_index:
            self.index_ivf = self.faiss.extract_index_ivf(self.index)
            self.index_ivf.metric_type = self.faiss.METRIC_INNER_PRODUCT
            self.nlist = self.index_ivf.nlist
            self.index_ivf.verbose = True
            self.downcast_quantizer = self.faiss.downcast_index(
                self.index_ivf.quantizer
            )
            self.downcast_quantizer.verbose = True
            self.downcast_quantizer.metric_type = self.faiss.METRIC_INNER_PRODUCT
            if hasattr(self.downcast_quantizer, 'hnsw'):
                self.downcast_quantizer.hnsw.efSearch = opt['hnsw_ef_search']
                self.downcast_quantizer.hnsw.efConstruction = opt[
                    'hnsw_ef_construction'
                ]
                self.downcast_quantizer.hnsw.metric_type = (
                    self.faiss.METRIC_INNER_PRODUCT
                )

            self.setup_gpu_train()
            self.index.nprobe = opt['compressed_indexer_nprobe']

        self.nprobe = opt['compressed_indexer_nprobe']
        self.span = 5  # arbitrarily chosen, from prior evidence
        self.random = random.Random(42)

    def setup_gpu_train(self):
        """
        Setup training on the gpu.
        """
        if self.use_gpu_train:
            logging.warning('Will train index on GPU')
            try:
                clustering_index = self.faiss.index_cpu_to_all_gpus(
                    self.faiss.IndexFlatIP(self.index_ivf.d)
                )
                self.index.clustering_index = clustering_index
            except NameError:
                logging.warning('GPU training not supported; switching to CPU.')

    def train(self, vectors: List[torch.Tensor]):
        """
        Train clustering index on list of vectors.

        :param vectors:
            a list of tensors to train on
        """
        start = time.time()
        for i, vecs in enumerate(grouper(vectors, self.span, None)):
            vec = torch.cat([v for v in vecs if v is not None])
            # sample
            num_samples = (
                min(50 * self.nlist, vec.size(0))
                if hasattr(self, 'n_list')
                else vec.size(0) // 2
            )
            vec = vec[torch.LongTensor(random.sample(range(vec.size(0)), num_samples))]
            logging.info(
                f'Training data {i+1}/{math.ceil(len(vectors) / self.span)} of shape {vec.shape}'
            )
            self.index.train(vec.float().numpy())
            logging.info(
                f'{time.time() - start:.2f}s Elapsed: training complete for {i+1}'
            )

    def add(self, vectors: List[torch.Tensor]):
        """
        Add vectors to index, using the CPU.

        :param vectors:
            vectors to add.
        """
        start = time.time()
        for i, vecs in enumerate(grouper(vectors, self.span, None)):
            vec = torch.cat([v for v in vecs if v is not None])
            logging.info(
                f'Adding data {(i+1)}/{math.ceil(len(vectors) / self.span)} of shape: {vec.shape}'
            )
            self.index.add(vec.float().numpy())
            logging.info(f'{time.time() - start}s Elapsed: adding complete for {i+1}')

    def index_data(self, data: List[torch.Tensor]):
        """
        Index data.

        :param data:
            list of (db_id, np.vector) tuples
        """
        start = time.time()
        assert isinstance(data, list)
        logging.info(f'Indexing {sum(v.size(0) for v in data)} vectors')
        # First, train
        self.train(data)

        # then, Add
        self.add(data)
        logging.info(f'Indexing complete; total time elapsed: {time.time() - start}')

    def deserialize_from(self, file: str, emb_path: Optional[str] = None):
        """
        If we're using an IVF index, we need to reset the nprobe parameter.

        `Make Direct Map` allows us to reconstruct vectors as well.
        """
        super().deserialize_from(file, emb_path)
        if self.is_ivf_index:
            self.index_ivf = self.faiss.extract_index_ivf(self.index)
            self.index_ivf.make_direct_map()
            self.index_ivf.nprobe = self.nprobe
            self.index.nprobe = self.nprobe


def indexer_factory(opt: Opt) -> BaseIndexer:
    """
    Build indexer.

    :param opt:
        Options

    :return indexer:
        return build indexer, according to options
    """
    if opt['indexer_type'] == 'compressed':
        if opt['path_to_index'] == WIKIPEDIA_EXACT_INDEX:
            logging.warning(
                f'Changing index path to compressed index: {WIKIPEDIA_COMPRESSED_INDEX}'
            )
            opt['path_to_index'] = modelzoo_path(
                opt['datapath'], WIKIPEDIA_COMPRESSED_INDEX
            )
        indexer = CompressedIndexer(opt)
    elif opt['indexer_type'] == 'exact':
        if opt['path_to_index'] == WIKIPEDIA_COMPRESSED_INDEX:
            logging.warning(
                f'Changing index path to exact index: {WIKIPEDIA_EXACT_INDEX}'
            )
            opt['path_to_index'] = modelzoo_path(opt['datapath'], WIKIPEDIA_EXACT_INDEX)
        indexer = DenseHNSWFlatIndexer(opt)
    else:
        raise ValueError(f"Unsupported indexer type: {opt['indexer_type']}")

    return indexer



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Opt is the system for passing around options throughout ParlAI.
"""

from __future__ import annotations

import copy
import json
import pickle
import traceback
import os
import pkg_resources
import parlai.utils.logging as logging

from typing import List

from parlai.utils.io import PathManager

# these keys are automatically removed upon save. This is a rather blunt hammer.
# It's preferred you indicate this at option definition time.
__AUTOCLEAN_KEYS__: List[str] = [
    "override",
    "batchindex",
    "download_path",
    "datapath",
    "verbose",
    # we don't save interactive mode or load from checkpoint, it's only decided by scripts or CLI
    "interactive_mode",
    "load_from_checkpoint",
]


class Opt(dict):
    """
    Class for tracking options.

    Functions like a dict, but allows us to track the history of arguments as they are
    set.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.history = []
        self.deepcopies = []

    def __setitem__(self, key, val):
        loc = traceback.format_stack(limit=2)[-2]
        self.history.append((key, val, loc))
        super().__setitem__(key, val)

    def __getstate__(self):
        return (self.history, self.deepcopies, dict(self))

    def __setstate__(self, state):
        self.history, self.deepcopies, data = state
        self.update(data)

    def __reduce__(self):
        return (Opt, (), self.__getstate__())

    def __deepcopy__(self, memo):
        """
        Override deepcopy so that history is copied over to new object.
        """
        # track location of deepcopy
        loc = traceback.format_stack(limit=3)[-3]
        self.deepcopies.append(loc)
        # copy all our children
        memo = Opt({k: copy.deepcopy(v) for k, v in self.items()})
        # deepcopy the history. history is only tuples, so we can do it shallow
        memo.history = copy.copy(self.history)
        # deepcopy the list of deepcopies. also shallow bc only strings
        memo.deepcopies = copy.copy(self.deepcopies)
        return memo

    def display_deepcopies(self):
        """
        Display all deepcopies.
        """
        if len(self.deepcopies) == 0:
            return 'No deepcopies performed on this opt.'
        return '\n'.join(f'{i}. {loc}' for i, loc in enumerate(self.deepcopies, 1))

    def display_history(self, key):
        """
        Display the history for an item in the dict.
        """
        changes = []
        i = 0
        for key_, val, loc in self.history:
            if key != key_:
                continue
            i += 1
            changes.append(f'{i}. {key} was set to {val} at:\n{loc}')
        if changes:
            return '\n'.join(changes)
        else:
            return f'No history for {key}'

    def save(self, filename: str) -> None:
        """
        Save the opt to disk.

        Attempts to 'clean up' any residual values automatically.
        """
        # start with a shallow copy
        dct = dict(self)

        # clean up some things we probably don't want to save
        for key in __AUTOCLEAN_KEYS__:
            if key in dct:
                del dct[key]

        with PathManager.open(filename, 'w', encoding='utf-8') as f:
            json.dump(dct, fp=f, indent=4)
            # extra newline for convenience of working with jq
            f.write('\n')

    @classmethod
    def load(cls, optfile: str) -> Opt:
        """
        Load an Opt from disk.
        """
        try:
            # try json first
            with PathManager.open(optfile, 'r', encoding='utf-8') as t_handle:
                dct = json.load(t_handle)
        except UnicodeDecodeError:
            # oops it's pickled
            with PathManager.open(optfile, 'rb') as b_handle:
                dct = pickle.load(b_handle)
        for key in __AUTOCLEAN_KEYS__:
            if key in dct:
                del dct[key]
        return cls(dct)

    @classmethod
    def load_init(cls, optfile: str) -> Opt:
        """
        Like load, but also looks in opt_presets folders.

        optfile may also be a comma-separated list of multiple presets/files.
        """
        if "," in optfile:
            # load and combine each of the individual files
            new_opt = cls()
            for subopt in optfile.split(","):
                new_opt.update(cls.load_init(subopt))
            return new_opt

        oa_filename = os.path.join("opt_presets", optfile + ".opt")
        user_filename = os.path.join(os.path.expanduser(f"~/.parlai"), oa_filename)
        if PathManager.exists(optfile):
            return cls.load(optfile)
        elif PathManager.exists(user_filename):
            # use a user's custom opt preset
            return cls.load(user_filename)
        else:
            # Maybe a bundled opt preset
            for root in ['parlai', 'parlai_internal', 'parlai_fb']:
                try:
                    if pkg_resources.resource_exists(root, oa_filename):
                        return cls.load(
                            pkg_resources.resource_filename(root, oa_filename)
                        )
                except ModuleNotFoundError:
                    continue

        # made it through without a return path so raise the error
        raise FileNotFoundError(
            f"Could not find filename '{optfile} or opt preset '{optfile}.opt'. "
            "Please check https://parl.ai/docs/opt_presets.html for a list "
            "of available opt presets."
        )

    def log(self, header="Opt"):
        from parlai.core.params import print_git_commit

        logging.info(header + ":")
        for key in sorted(self.keys()):
            valstr = str(self[key])
            if valstr.replace(" ", "").replace("\n", "") != valstr:
                # show newlines as escaped keys, whitespace with quotes, etc
                valstr = repr(valstr)
            logging.info(f"    {key}: {valstr}")
        print_git_commit()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = 'wizard_of_internet'


class TestApprenticeTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:ApprenticeDialogTeacher'


class TestWizardGoldKnowledgeTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:WizardDialogGoldKnowledgeTeacher'


class TestSearchQueryTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:SearchQueryTeacher'


class TestGoldKnowledgeTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:GoldKnowledgeTeacher'


class TestGoldDocsTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:GoldDocsTeacher'


class TestGoldDocTitlesTeacher(AutoTeacherTest):
    task = 'wizard_of_internet:GoldDocTitlesTeacher'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = "superglue"


class TestBoolqTeacher(AutoTeacherTest):
    task = "superglue:boolq"


class TestCbTeacher(AutoTeacherTest):
    task = "superglue:cb"


class TestCopaTeacher(AutoTeacherTest):
    task = "superglue:copa"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import json
import os
import random
import shutil
from tqdm import tqdm

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
from parlai.utils.logging import logger


RESOURCES = [
    DownloadableFile(
        'https://github.com/festvox/datasets-CMU_DoG/archive/618a14f27546165859305649aa84e6ac8710bb63.zip',
        'cmu_dog.zip',
        'f8ba8820cf86ee1c196b237b0cde80edba940e4ddea28c582830f6d098b3c769',
    )
]

UNZIPPED_PARENT_DIR = 'datasets-CMU_DoG-618a14f27546165859305649aa84e6ac8710bb63'


def build(opt):
    dpath = os.path.join(opt['datapath'], 'cmu_dog')
    version = '1.2'
    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        move_unzipped_files_up(dpath)
        consolidate_wiki_data(dpath)
        consolidate_convos(dpath)
        build_deduped_split(dpath)
        split_into_seen_unseen(dpath)

        build_data.mark_done(dpath, version)


def move_unzipped_files_up(dpath: str):
    unzipped_path = os.path.join(dpath, UNZIPPED_PARENT_DIR)
    for f in os.listdir(unzipped_path):
        shutil.move(os.path.join(unzipped_path, f), dpath)
    shutil.rmtree(unzipped_path)


def consolidate_wiki_data(dpath: str):
    all_articles = {}
    src_dir = os.path.join(dpath, 'WikiData')
    for f_name in tqdm(os.listdir(src_dir)):
        with open(os.path.join(src_dir, f_name)) as f:
            wiki_page = json.load(f)
        idx = wiki_page['wikiDocumentIdx']
        all_articles[idx] = wiki_page
    dest_path = os.path.join(dpath, 'wiki_data.json')
    with open(dest_path, 'w') as d:
        json.dump(all_articles, d, indent=2)
    shutil.rmtree(src_dir)


def consolidate_convos(dpath: str):
    os.makedirs(os.path.join(dpath, 'conversations'), exist_ok=True)
    for split in ['train', 'valid', 'test']:
        consolidate_convo_split(dpath, split)


def consolidate_convo_split(dpath: str, split: str):
    all_convos = {}
    src_dir = os.path.join(dpath, 'Conversations', split)
    for f_name in tqdm(os.listdir(src_dir)):
        with open(os.path.join(src_dir, f_name)) as f:
            convo = json.load(f)
        cid = f_name.split('.')[0]
        all_convos[cid] = convo
    dest_path = os.path.join(dpath, 'conversations', f"{split}.json")
    with open(dest_path, 'w') as dest:
        json.dump(all_convos, dest, indent=2)
    shutil.rmtree(src_dir)


def build_deduped_split(dpath: str):
    """
    Original CMU-DoG has 110 ids that are used in multiple of train/valid/test.

    Get rid of the duplication.
    """
    cdir = os.path.join(dpath, "conversations")
    data = {}
    for fold in ["test", "valid", "train"]:
        fpath = os.path.join(cdir, f"{fold}.json")
        with PathManager.open(fpath) as f:
            data[fold] = json.load(f)

    train_len = len(data["train"])
    valid_len = len(data["valid"])
    test_len = len(data["test"])
    logger.info(
        f"Converation count with duplicates: train-{train_len}, valid-{valid_len}, test-{test_len}"
    )

    train_valid = set(data["train"].keys()) & set(data["valid"].keys())
    train_test = set(data["train"].keys()) & set(data["test"].keys())
    valid_test = set(data["valid"].keys()) & set(data["test"].keys())

    for key in train_valid:
        data["train"].pop(key)
    for key in train_test:
        data["train"].pop(key)
    for key in valid_test:
        data["test"].pop(key)

    train_len = len(data["train"])
    valid_len = len(data["valid"])
    test_len = len(data["test"])
    logger.info(
        f"Converation count without duplicates: train-{train_len}, valid-{valid_len}, test-{test_len}"
    )

    for fold in ["test", "valid", "train"]:
        fpath = os.path.join(cdir, f"{fold}_deduped.json")
        with PathManager.open(fpath, "w+") as f:
            json.dump(data[fold], f, indent=2)


def split_into_seen_unseen(dpath: str):
    """
    Following WoW, we have overlap in train, valid, and test seen but none in test
    unseen. Do an 80:10:5:5 split between train, valid, test_seen, test_unseen or as
    close to it.

    ~205 documents for test_unseen to do this, and movies 1 and 3 have 90 and 117
    movies, respectively, which is about that
    """
    random.seed(42)
    cdir = os.path.join(dpath, "conversations")
    new = {"train": {}, "valid": {}, "test_seen": {}, "test_unseen": {}}
    for fold in ["test", "valid", "train"]:
        with PathManager.open(os.path.join(cdir, f"{fold}_deduped.json")) as f:
            data = json.load(f)
        for k, v in data.items():
            if v["wikiDocumentIdx"] == 1 or v["wikiDocumentIdx"] == 3:
                new["test_unseen"][k] = v
            else:
                rand = random.randint(1, 95)
                if rand <= 80:
                    new["train"][k] = v
                elif rand <= 90:
                    new["valid"][k] = v
                else:
                    new["test_seen"][k] = v

    for fold in new:
        with PathManager.open(
            os.path.join(cdir, f"{fold}_split_seen_unseen.json"), "w+"
        ) as f:
            json.dump(new[fold], f, indent=2)
        c_cnt = len(new[fold])
        logger.info(f"Seen/unseen {fold} conversation count: {c_cnt}")



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
T5: Exploring the Limits of Transfer Learning with a Unified Text-to-Text Transformer

See <https://arxiv.org/abs/1910.10683>

The T5 agent can be instantiated as simply `-m t5`
"""
import torch
from typing import Optional, Dict, Any, Tuple
from transformers import T5ForConditionalGeneration

try:
    from transformers.models.t5.modeling_t5 import T5Stack
except ModuleNotFoundError:
    # Prior versions of transformers package do not have T5Stack
    T5Stack = object

from parlai.agents.hugging_face.hugging_face import HF_VERSION
from parlai.agents.hugging_face.dict import T5DictionaryAgent

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import Batch, TorchAgent
from parlai.core.torch_generator_agent import TorchGeneratorAgent, TorchGeneratorModel


def build_t5(opt: Opt) -> T5ForConditionalGeneration:
    if not HF_VERSION >= 4.3:
        raise RuntimeError('Must use transformers package >= 4.3 to use t5')
    return T5ForConditionalGeneration.from_pretrained(
        opt['t5_model_arch'], dropout_rate=opt['t5_dropout']
    )


def set_device(func):
    """
    Decorator for setting device.

    HF's model parallel uses `torch.cuda.set_device`, which does not vibe well with
    ParlAI.
    """

    def wrap(*args, **kwargs):
        if torch.cuda.is_available():
            torch.cuda.set_device('cuda:0')
        ret = func(*args, **kwargs)
        if torch.cuda.is_available():
            torch.cuda.set_device('cuda:0')
        return ret

    return wrap


class T5Agent(TorchGeneratorAgent):
    """
    T5 Agent.

    Relies on the T5 model implemented in huggingface
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        group = parser.add_argument_group('T5 Args')
        group.add_argument(
            '--t5-model-arch',
            type=str,
            default='t5-base',
            choices=["t5-small", "t5-base", "t5-large", "t5-3b", "t5-11b"],
        )
        group.add_argument(
            '--t5-model-parallel',
            type='bool',
            default=False,
            help='use HF model parallel',
        )
        group.add_argument(
            '--t5-dropout', type=float, default=0.0, help='Dropout for T5'
        )
        group.add_argument(
            '--t5-generation-config',
            type=str,
            default=None,
            choices=[
                'summarization',
                'translation_en_to_de',
                'translation_en_to_fr',
                'translation_en_to_ro',
            ],
            help='Task specific generation config for T5',
        )
        return parser

    def build_model(self) -> 'ParlaiT5Model':
        """
        Build and return model.
        """
        model = ParlaiT5Model(self.opt, self.dict)
        if self.opt['t5_model_parallel']:
            model.t5.parallelize()
        return model

    def build_dictionary(self):
        """
        Overrides TorchAgent.build_dictionary to use t5 dict.
        """
        return T5DictionaryAgent(self.opt)

    def vectorize(self, *args, **kwargs):
        """
        Override vectorize for T5.

        T5 dict already adds the end token.
        """
        kwargs['add_start'] = False  # model does this in module code
        kwargs['add_end'] = False  # T5 tokenizer takes care of this
        return TorchAgent.vectorize(self, *args, **kwargs)

    def observe(self, observation):
        """
        Override to include prefix, if necessary.
        """
        if self.opt['t5_generation_config'] is not None and 'text' in observation:
            config = TASK_CONFIGS[self.opt['t5_generation_config']]
            try:
                observation.force_set('text', config['prefix'] + observation['text'])
            except AttributeError:
                observation['text'] = config['prefix'] + observation['text']

        return super().observe(observation)

    def _generate(
        self,
        batch: Batch,
        beam_size: int,
        max_ts: int,
        prefix_tokens: Optional[torch.LongTensor] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ):
        """
        Generate an output with beam search.

        Use HF's built-in generation to perform beam search.
        """
        bad_words_ids = None
        if self.beam_block_list is not None:
            bad_words_ids = [
                gram for _, ngram in self.beam_block_list.items() for gram in ngram
            ]

        method = self.opt.get('inference', 'greedy')

        generation_params = {
            'input_ids': batch.text_vec,
            'max_length': max_ts,
            'min_length': self.beam_min_length,
            'do_sample': self.opt['inference'] in ['topk', 'topp'],
            'early_stopping': None,
            'num_beams': beam_size,
            'temperature': self.temperature,
            'top_k': self.opt['topk'] if method in ['topk', 'delayedbeam'] else None,
            'top_p': self.opt['topp'] if method == 'nucleus' else None,
            'repetition_penalty': None,
            'bad_words_ids': bad_words_ids if bad_words_ids else None,
            'bos_token_id': self.START_IDX,
            'pad_token_id': self.NULL_IDX,
            'eos_token_id': self.END_IDX,
            'length_penalty': self.opt['beam_length_penalty'],
            'no_repeat_ngram_size': self.beam_block_ngram,
            'num_return_sequences': None,
            'attention_mask': batch.text_vec != self.NULL_IDX,
            'decoder_start_token_id': self.NULL_IDX,
        }

        if self.opt['t5_generation_config']:
            config = TASK_CONFIGS[self.opt['t5_generation_config']]
            config.pop('prefix', None)
            generation_params.update(config)
        if overrides:
            generation_params.update(overrides)

        outputs = self.model.t5.generate(**generation_params)
        outputs = [(outputs[i], 0) for i in range(outputs.size(0))]
        return outputs, []


##############
# T5 Modules #
##############


class ParlaiT5Encoder(torch.nn.Module):
    def __init__(self, opt: Opt, encoder: T5Stack, padding_idx: Optional[int] = None):
        super().__init__()
        self.stack = encoder
        self.padding_idx = padding_idx
        self.paralleled = not opt[
            't5_model_parallel'
        ]  # need to parallel in forward; bug in HF

    @set_device
    def forward(
        self,
        input: torch.LongTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[torch.Tensor, torch.BoolTensor]:
        """
        Forward pass.

        :param LongTensor[batch,seqlen] input:
            The input IDs
        :param LongTensor[batch,seqlen] positions:
            Positions for input IDs
        :param LongTensor[batch,seqlen] segments:
            If provided, additionally adds ``segments`` as extra embedding features.
        """
        if not self.paralleled:
            self.stack.parallelize()
        mask = input != self.padding_idx
        outputs = self.stack(input, attention_mask=mask, output_hidden_states=False)
        for k in outputs:
            if torch.is_tensor(outputs[k]):
                outputs[k] = outputs[k].to(input.device)
        return outputs[0], mask


class ParlaiT5Decoder(torch.nn.Module):
    def __init__(self, opt: Opt, decoder: T5Stack, padding_idx: Optional[int] = None):
        super().__init__()
        self.stack = decoder
        self.padding_idx = padding_idx
        self.paralleled = not opt[
            't5_model_parallel'
        ]  # need to parallel in forward; bug in HF

    @set_device
    def forward(
        self, input: torch.LongTensor, encoder_state: Tuple[Any], incr_state=None
    ):
        """
        Forward pass.

        :param LongTensor[batch,seqlen] input:
            The decoder inputs (partial or full decoded token IDs).
        :param encoder_state:
            Output from the encoder module forward pass.
        :param incr_state:
            The incremental state: a dictionary whose keys index the layers and whose
            values contain the incremental state for each layer.
        """
        if not self.paralleled:
            self.stack.parallelize()
        encoder_output, encoder_mask = encoder_state

        mask = input != self.padding_idx
        mask[:, 0] = True  # first token is pad

        outputs = self.stack(
            input_ids=input,
            attention_mask=mask,
            encoder_hidden_states=encoder_output.to(input.device),
            encoder_attention_mask=encoder_mask.to(input.device),
        )
        return outputs[0].to(input.device), incr_state


class ParlaiT5Model(TorchGeneratorModel):
    """
    Wrap T5 in ParlAI.
    """

    def __init__(self, opt, dictionary):
        self.pad_idx = dictionary[dictionary.null_token]
        self.start_idx = self.pad_idx
        self.end_idx = dictionary[dictionary.end_token]
        super().__init__(self.pad_idx, self.start_idx, self.end_idx)
        self.t5 = build_t5(opt)
        self.encoder = ParlaiT5Encoder(opt, self.t5.get_encoder(), self.pad_idx)
        self.decoder = ParlaiT5Decoder(opt, self.t5.get_decoder(), self.pad_idx)

    @set_device
    def _get_initial_forced_decoder_input(self, bsz: int, inputs: torch.LongTensor):
        """
        Return initial input to the decoder.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode

        :return initial_input:
            initial input for the decoder.
        """
        inputs = torch.cat([self.START.detach().expand(bsz, 1), inputs], 1)
        return inputs

    @set_device
    def reorder_encoder_states(self, encoder_states, indices):
        """
        Reorder the encoder states.

        See ``TorchGeneratorModel.reorder_encoder_states`` for a description.
        """
        enc, mask = encoder_states
        if not torch.is_tensor(indices):
            indices = torch.LongTensor(indices).to(enc.device)
        enc = torch.index_select(enc, 0, indices)
        mask = torch.index_select(mask, 0, indices)
        return enc, mask

    def reorder_decoder_incremental_state(
        self, incremental_state: Dict[int, dict], inds: torch.Tensor
    ) -> Dict[int, dict]:
        """
        Not *quite* sure how to reconcile this with HF.
        """
        return {}

    @set_device
    def output(self, tensor):
        """
        Compute output logits.
        """
        # Taken directly from HuggingFace
        # Rescale output before projecting on vocab
        # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
        tensor = tensor * (self.t5.model_dim ** -0.5)
        lm_logits = self.t5.lm_head(tensor)
        return lm_logits


###########################################################################
# Task-specific generation configs for T5.                                #
# Taken from HF: https://huggingface.co/t5-small/resolve/main/config.json #
###########################################################################

TASK_CONFIGS = {
    "summarization": {
        "early_stopping": True,
        "length_penalty": 2.0,
        "max_length": 200,
        "min_length": 30,
        "no_repeat_ngram_size": 3,
        "num_beams": 4,
        "prefix": "summarize: ",
    },
    "translation_en_to_de": {
        "early_stopping": True,
        "max_length": 300,
        "num_beams": 4,
        "prefix": "translate English to German: ",
    },
    "translation_en_to_fr": {
        "early_stopping": True,
        "max_length": 300,
        "num_beams": 4,
        "prefix": "translate English to French: ",
    },
    "translation_en_to_ro": {
        "early_stopping": True,
        "max_length": 300,
        "num_beams": 4,
        "prefix": "translate English to Romanian: ",
    },
}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Pre-trained DPR Model.
"""
import gzip
import os
import os.path

from parlai.core.build_data import built, download as download_path, mark_done
import parlai.utils.logging as logger

path = 'https://dl.fbaipublicfiles.com/dpr/wikipedia_split/psgs_w100.tsv.gz'


def download(datapath):
    dpath = os.path.join(datapath, 'models/hallucination/wiki_passages')
    fname = 'psgs_w100.tsv.gz'
    gzip_file = os.path.join(dpath, fname)
    new_file = os.path.join(dpath, fname.replace('.gz', ''))
    version = 'v1.0'
    if not built(dpath, version):
        os.makedirs(dpath)
        download_path(path, dpath, fname)
        input = gzip.GzipFile(gzip_file, "rb")
        s = input.read()
        input.close()
        output = open(new_file, "wb")
        output.write(s)
        output.close()
        logger.info(f" Saved to {new_file}")
        mark_done(dpath, version)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

LAYER_NORM_EPS = 1e-5  # Epsilon for layer norm.



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


INITIAL_DATA_TO_COMPLETE = [
    'valid_0',
    'valid_1',
    'valid_10',
    'valid_100',
    'valid_101',
    'valid_102',
    'valid_103',
    'valid_105',
    'valid_106',
    'valid_107',
    'valid_108',
    'valid_109',
    'valid_11',
    'valid_110',
    'valid_111',
    'valid_112',
    'valid_115',
    'valid_116',
    'valid_117',
    'valid_119',
    'valid_12',
    'valid_120',
    'valid_121',
    'valid_122',
    'valid_123',
    'valid_124',
    'valid_125',
    'valid_13',
    'valid_130',
    'valid_131',
    'valid_133',
    'valid_136',
    'valid_138',
    'valid_139',
    'valid_14',
    'valid_140',
    'valid_141',
    'valid_143',
    'valid_144',
    'valid_145',
    'valid_146',
    'valid_147',
    'valid_148',
    'valid_15',
    'valid_152',
    'valid_153',
    'valid_154',
    'valid_155',
    'valid_156',
    'valid_158',
    'valid_160',
    'valid_162',
    'valid_163',
    'valid_166',
    'valid_169',
    'valid_17',
    'valid_171',
    'valid_172',
    'valid_174',
    'valid_175',
    'valid_176',
    'valid_177',
    'valid_178',
    'valid_18',
    'valid_181',
    'valid_182',
    'valid_184',
    'valid_187',
    'valid_19',
    'valid_190',
    'valid_191',
    'valid_192',
    'valid_193',
    'valid_194',
    'valid_196',
    'valid_2',
    'valid_20',
    'valid_202',
    'valid_203',
    'valid_205',
    'valid_206',
    'valid_207',
    'valid_208',
    'valid_212',
    'valid_214',
    'valid_215',
    'valid_216',
    'valid_217',
    'valid_219',
    'valid_223',
    'valid_225',
    'valid_227',
    'valid_228',
    'valid_23',
    'valid_230',
    'valid_231',
    'valid_232',
    'valid_233',
    'valid_234',
    'valid_236',
]

COMMON_CONFIG = {
    'task': 'msc:SessionBaseMsc',
    'num_examples': -1,
    'label_speaker_id': 'their',
    'session_id': 4,
    'datatype': 'valid',
}

MODEL_OPT = {
    'BST90M': {
        'previous_persona_type': 'none',
        'num_previous_sessions_msg': 10,
        'include_time_gap': False,
    }
}

UI_OPT = {'BST90M': {'previous_persona_type': 'both', 'include_time_gap': False}}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np
from typing import Iterable, List
from nltk.stem import PorterStemmer
from parlai.crowdsourcing.utils.acceptability import (
    AcceptabilityChecker,
    normalize_answer,
)
import parlai.utils.logging as logging


# Bad persona violations
PERSONA_REPEATS_PROMPT = 'repeated the prompt text'
ASKED_WIZARD_QUESTION = 'asked wizard in the persona details'
COPIED_EXTENDED_PERSONA = 'extended persona copies the main persona'
GENERIC_EXTENDED_PERSONA = 'extended persona is generic'

QUESTION_PHRASE = 'what is your'

# Wizard knowledge violations
DEFAULT_KNOWLEDGE_OVERLAP_THRESHOLD = 0.05

POOR_SEARCH_QUERIES = 'poor search queries'
IRRELEVANT_SEARCH__QUERIES = 'irrelevant search terms'
NOT_ENOUGH_SEARCH = 'not enough selected knowledge sources'
SELECTED_SHORT_PIECES = 'short knowledge pieces selected.'
LOW_KNOWLEDGE_OVERLAP = 'low knowledge overlap'


def tokenize_text(text, stemmer, as_set=True):
    text = normalize_answer(text)
    tokens = [stemmer.stem(word) for word in text.split(' ')]
    if as_set:
        tokens = set(tokens)
    return tokens


def overlap_ratios(a: set, b: set) -> float:
    """
    Calculates the Jacard distance between two sets.
    """
    overlap = a.intersection(b)
    union = a.union(b)
    return len(overlap) / (len(union) + 0.001)


def is_valid_agent_chat_message(message, agent_id):
    return (
        message.get('text')
        and message.get('id') == agent_id
        and not message.get('is_search_query', False)
    )


def bad_persona(persona, stemmer):
    """
    Check for poor persona selection by apprentice.
    """
    persona_parts = persona.split('\n')

    # It is not from the persona selection ones (personas used during the pilot).
    if not (
        len(persona_parts) == 2
        or (len(persona_parts) == 3 and 'I live in ' in persona_parts[0])
    ):
        logging.warning(f'Old fashioned persona: {persona}')
        return

    # Removing the location ('I live in X') part
    if len(persona_parts) == 3:
        persona_parts = persona_parts[1:]

    main_pers, ext_pers = [p.lower() for p in persona_parts]

    violations = []

    # Bad main persona response
    if main_pers.startswith('My favorite '):
        for phrase in ('i like', 'my favorite'):
            persona_core = main_pers
            # Remove the original My favorite
            persona_core = main_pers[len('My favorite ') :]
            if phrase in persona_core.lower():
                violations.append(PERSONA_REPEATS_PROMPT)
                break

    # Extended persona that asks questions
    for phrase in (QUESTION_PHRASE,):
        if phrase in ext_pers:
            violations.append(ASKED_WIZARD_QUESTION)

    # Extended persona that mostly repeats the main persona
    main_pers_tokens = tokenize_text(main_pers, stemmer)
    ext_pers_tokens = tokenize_text(ext_pers, stemmer)
    if len(ext_pers_tokens.difference(main_pers_tokens)) < 2:
        violations.append(COPIED_EXTENDED_PERSONA)

    # Use of non-generic words in persona.
    common_phrases = ('i', 'it', 'like', 'very', 'much', 'favorite', 'is', 'am')
    tokens = [w.strip() for w in ext_pers.split(' ') if w]
    ext_useful_words = [t for t in tokens if t not in common_phrases]
    if len(tokens) > 4 and len(ext_useful_words) < 2:
        violations.append(GENERIC_EXTENDED_PERSONA)

    return violations


def poor_knowledge_selection(messages, persona, stemmer, knwldg_ovlp_thrshld):
    """
    Check for poor search and knowledge selection by wizard.
    """
    # Collecting search and knowledge selections
    search_terms = []
    selected_knowledge = []
    message_history_tokens = tokenize_text(persona, stemmer)

    n_search_query_not_in_history = 0
    for msg in messages:
        if msg.get('text', None):
            message_history_tokens = message_history_tokens.union(
                tokenize_text(msg['text'], stemmer)
            )

        if msg['id'] != 'Wizard':
            continue

        selections = msg.get('task_data', {}).get('selected_text_candaidtes')
        if not selections or selections[0][0]:
            continue

        search_query = msg['task_data']['search_query']
        search_terms.append(search_query)
        if message_history_tokens.isdisjoint(tokenize_text(search_query, stemmer)):
            n_search_query_not_in_history += 1

        selected_parts = []
        for doc_id in range(1, len(selections)):
            doc_selections = selections[doc_id]
            for sentence_id in range(len(doc_selections)):
                if doc_selections[sentence_id]:
                    selected_parts.append(
                        msg['task_data']['text_candidates'][doc_id - 1]['content'][
                            sentence_id
                        ]
                    )

        selected_knowledge.append(
            {'text': msg['text'], 'knowledge': ' '.join(selected_parts)}
        )

    knowledge_length = []
    knowledge_overlaps = []
    for knwldg in selected_knowledge:
        knowledge_tokens = tokenize_text(knwldg['knowledge'], stemmer)
        knowledge_length.append(len(knowledge_tokens))

        response_tokens = tokenize_text(knwldg['text'], stemmer)
        knowledge_overlaps.append(overlap_ratios(knowledge_tokens, response_tokens))

    violations = []

    # Repeated the same search queries
    if len(search_terms) - len(set(search_terms)) > 3:
        violations.append(POOR_SEARCH_QUERIES)

    # Search doesn't have overlap with message history
    if n_search_query_not_in_history > 2:
        violations.append(IRRELEVANT_SEARCH__QUERIES)

    # No selection
    if not knowledge_length:
        violations.append(NOT_ENOUGH_SEARCH)

    # Only selecting short sentences
    if np.average(knowledge_length) < 5:
        violations.append(SELECTED_SHORT_PIECES)

    # Small overlap between response and the selected knowledge parts
    knowledge_overlap_avg = np.average(knowledge_overlaps)
    if knowledge_overlap_avg < knwldg_ovlp_thrshld:
        violations.append(f'{LOW_KNOWLEDGE_OVERLAP} ({knowledge_overlap_avg})')

    return violations


class WizardOfInternetAcceptabilityChecker(AcceptabilityChecker):
    """
    ParlAI general acceptabilty checker customized for the wizard of internet.
    """

    def __init__(self):
        self.knowledge_overlap_threshold = DEFAULT_KNOWLEDGE_OVERLAP_THRESHOLD
        self.post_stemmer = PorterStemmer()
        super().__init__()

    def check_messages(
        self,
        agent_id: str,
        persona: str,
        messages: List[str],
        is_worker_0: bool,
        violation_types: Iterable[str] = (),
    ) -> str:
        violations = []
        general_chat_violations = super().check_messages(
            self.get_conversation_messages(messages, agent_id),
            is_worker_0,
            violation_types,
        )
        if general_chat_violations:
            violations.extend(general_chat_violations.split(','))

        if agent_id == 'Apprentice':
            persona_violations = bad_persona(persona, self.post_stemmer)
            if persona_violations:
                violations.extend(persona_violations)

        if agent_id == 'Wizard':
            knowledge_violations = poor_knowledge_selection(
                messages, persona, self.post_stemmer, self.knowledge_overlap_threshold
            )
            if knowledge_violations:
                violations.extend(knowledge_violations)

        return ','.join(violations)

    def get_conversation_messages(self, agent_messages, agent_id):
        return [
            msg['text']
            for msg in agent_messages
            if is_valid_agent_chat_message(msg, agent_id)
        ]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BB2, 3B.
"""
from parlai.core.build_data import built, download_models, get_model_dir
import os
import os.path


def download(datapath):
    ddir = os.path.join(get_model_dir(datapath), 'blenderbot2')
    model_type = 'blenderbot2_3B'
    version = 'v1.0'
    if not built(os.path.join(ddir, model_type), version):
        opt = {'datapath': datapath, 'model_type': model_type}
        fnames = ['model.tgz']
        download_models(
            opt, fnames, 'blenderbot2', version=version, use_model_type=True
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Create a FAISS Index with a series of dense embeddings.
"""
import os
import random
import torch
from typing import List

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript
import parlai.utils.logging as logging

from parlai.agents.rag.rag import RagAgent
from parlai.agents.rag.indexers import indexer_factory, CompressedIndexer


class Indexer(ParlaiScript):
    """
    Index Dense Embeddings.
    """

    @classmethod
    def setup_args(cls):
        """
        Setup args.
        """
        parser = ParlaiParser(True, True, 'Index Dense Embs')
        parser.add_argument(
            '--embeddings-dir', type=str, help='directory of embeddings'
        )
        parser.add_argument(
            '--embeddings-name', type=str, default='', help='name of emb part'
        )
        parser.add_argument(
            '--partition-index',
            type='bool',
            default=False,
            help='specify True to partition indexing per file (useful when all files do not fit into memory)',
        )
        parser.add_argument(
            '--save-index-dir',
            type=str,
            help='directory in which to save index',
            default=None,
        )
        parser.add_argument(
            '--num-shards',
            type=int,
            default=1,
            help='how many workers to use to split up the work',
        )
        parser.add_argument(
            '--shard-id',
            type=int,
            help='shard id for this worker. should be between 0 and num_shards',
        )
        parser = RagAgent.add_cmdline_args(parser)
        parser.set_defaults(compressed_indexer_gpu_train=True)
        return parser

    def run(self):
        """
        Load dense embeddings and index with FAISS.
        """
        # create index
        index_dir = self.opt['embeddings_dir']
        embs_name = (
            f"{self.opt['embeddings_name']}_" if self.opt['embeddings_name'] else ''
        )
        num_parts = len(
            [
                f
                for f in os.listdir(index_dir)
                if f.endswith('.pt') and 'sample' not in f
            ]
        )
        input_files = [
            os.path.join(index_dir, f'{embs_name}{i}.pt') for i in range(num_parts)
        ]
        if self.opt['indexer_type'] == 'compressed':
            index_name = self.opt['compressed_indexer_factory'].replace(',', '__')
        elif self.opt['embeddings_name']:
            index_name = self.opt['embeddings_name']
        else:
            index_name = 'hnsw_flat'
        index_path = os.path.join(index_dir, index_name)

        if self.opt['save_index_dir']:
            index_path, index_name = os.path.split(index_path)
            index_path = os.path.join(self.opt['save_index_dir'], index_name)
            if not os.path.exists(self.opt['save_index_dir']):
                logging.info(f'Creating directory for file {index_path}')
                os.makedirs(self.opt['save_index_dir'])

        logging.info(f'index path: {index_path}')
        self.index_path = index_path

        self.index = indexer_factory(self.opt)
        if self.opt['indexer_type'] != 'exact':
            self.train_then_add(input_files)
        else:
            self.index_data(input_files)
        # save data
        self.index.serialize(index_path)

    def index_data(self, input_files: List[str], add_only: bool = False):
        """
        Index data.

        :param input_files:
            files to load.
        """
        all_docs = []
        for in_file in input_files:
            logging.info(f'Reading file {in_file}')
            docs = torch.load(in_file)
            if isinstance(docs, list):
                all_docs += docs
            else:
                all_docs.append(docs)

        self.index.index_data(all_docs)

    def train_then_add(self, input_files: List[str]):
        """
        First train data, then add it.

        If we're only training... then don't add!!
        """
        assert isinstance(self.index, CompressedIndexer)
        # Train
        random.seed(42)

        tensors = []
        for s in input_files:
            logging.info(f'Loading in file {s}')
            tensor = torch.load(s)
            if self.opt['partition_index']:
                self.index.train([tensor])
                self.index.add([tensor])
            else:
                tensors.append(tensor)

        if not self.opt['partition_index']:
            self.index.train([torch.cat(tensors)])
            self.index.add([torch.cat(tensors)])


if __name__ == "__main__":
    Indexer.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Wrappers for ParlAI models in the model zoo.

Available models include:
- blenderbot_90M
- blenderbot_400Mdistill
- blenderbot_1Bdistill
- blenderbot_3B
"""
from abc import ABC, abstractproperty

from parlai.core.agents import create_agent_from_model_file
from projects.safety_bench.utils.wrapper_loading import register_model_wrapper


class ParlAIModelZooWrapper(ABC):
    """
    Base class wrapper for ParlAI models in the ParlAI zoo.
    """

    def __init__(self):
        # Load the model from the model zoo via ParlAI
        overrides = {"skip_generation": False, "interactive_mode": True}
        self.model = create_agent_from_model_file(self.zoo_path, overrides)

    @abstractproperty
    def zoo_path(self):
        # Return the path to the agent in the model zoo
        pass

    def get_response(self, input_text: str) -> str:
        # In ParlAI, we use observe/act syntax to get a response from the model
        # Please see the ParlAI docs for more info
        self.model.observe({"text": input_text, "episode_done": True})
        response = self.model.act()

        return response.get("text")


@register_model_wrapper("blenderbot_90M")
class BlenderBot90MWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_90M/model"


@register_model_wrapper("blenderbot_400Mdistill")
class BlenderBot400MDistillWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_400Mdistill/model"


@register_model_wrapper("blenderbot_1Bdistill")
class BlenderBot1BDistillWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_1Bdistill/model"


@register_model_wrapper("blenderbot_3B")
class BlenderBot3BWrapper(ParlAIModelZooWrapper):
    @property
    def zoo_path(self):
        return "zoo:blender/blender_3B/model"
